package ru.dbotthepony.mc.otm.container;

import net.minecraft.world.item.ItemStack;
import net.minecraftforge.common.util.LazyOptional;
import net.minecraftforge.items.IItemHandler;

import javax.annotation.Nonnull;

public class MatteryContainerHandler implements IItemHandler {
	private final MatteryContainer container;
	private final MatteryContainerInsertValidator insert_validator;
	private final MatteryContainerExtractValidator extract_validator;

	MatteryContainerHandler(MatteryContainer container, MatteryContainerInsertValidator insert_validator, MatteryContainerExtractValidator extract_validator) {
		this.container = container;
		this.insert_validator = insert_validator;
		this.extract_validator = extract_validator;
	}

	MatteryContainerHandler(MatteryContainer container) {
		this(container, (slot, stack) -> true, (slot, amount, stack) -> true);
	}

	MatteryContainerHandler(MatteryContainer container, MatteryContainerInsertValidator insert_validator) {
		this(container, insert_validator, (slot, amount, stack) -> true);
	}

	private LazyOptional<IItemHandler> handler = LazyOptional.of(() -> this);

	public LazyOptional<IItemHandler> get() {
		return handler;
	}

	public void invalidate() {
		handler.invalidate();
	}

	public void revive() {
		handler = LazyOptional.of(() -> this);
	}

	@Override
	public int getSlots() {
		return container.getContainerSize();
	}

	@Nonnull
	@Override
	public ItemStack getStackInSlot(int slot) {
		return container.getItem(slot);
	}

	@Nonnull
	@Override
	public ItemStack insertItem(int slot, @Nonnull ItemStack stack, boolean simulate) {
		if (!insert_validator.apply(slot, stack))
			return stack;

		ItemStack self_stack = container.getItem(slot);

		if (self_stack.isEmpty()) {
			if (!simulate) {
				var copy = stack.copy();
				container.setChanged(slot, copy, ItemStack.EMPTY);
				container.setItem(slot, copy);
			}

			return ItemStack.EMPTY;
		} else if (self_stack.isStackable() && self_stack.getMaxStackSize() > self_stack.getCount() && ItemStack.isSameItemSameTags(self_stack, stack)) {
			int new_count = Math.min(self_stack.getMaxStackSize(), self_stack.getCount() + stack.getCount());
			int diff = new_count - self_stack.getCount();

			if (diff != 0) {
				if (!simulate) {
					ItemStack copy_self = self_stack.copy();
					self_stack.grow(diff);
					container.setChanged(slot, self_stack, copy_self);
				}

				ItemStack copy = stack.copy();
				copy.shrink(diff);
				return copy;
			}
		}

		return stack;
	}

	@Nonnull
	@Override
	public ItemStack extractItem(int slot, int amount, boolean simulate) {
		if (amount == 0)
			return ItemStack.EMPTY;

		if (amount < 0)
			throw new IllegalArgumentException("Can not extract negative amount of items");

		ItemStack self_stack = container.getItem(slot);

		if (self_stack.isEmpty())
			return ItemStack.EMPTY;

		if (!extract_validator.apply(slot, amount, self_stack))
			return ItemStack.EMPTY;

		int minimal = Math.min(amount, self_stack.getCount());
		ItemStack copy = self_stack.copy();
		copy.setCount(minimal);

		if (!simulate) {
			ItemStack copy_self = self_stack.copy();
			self_stack.shrink(minimal);
			container.setChanged(slot, self_stack, copy_self);
		}

		return copy;
	}

	@Override
	public int getSlotLimit(int slot) {
		return container.getMaxStackSize();
	}

	@Override
	public boolean isItemValid(int slot, @Nonnull ItemStack stack) {
		return insert_validator.apply(slot, stack);
	}

	MatteryContainerHandler cloneOf(MatteryContainer container) {
		return new MatteryContainerHandler(container, insert_validator, extract_validator);
	}
}
