package ru.dbotthepony.mc.otm.block.entity

import net.minecraft.core.BlockPos
import net.minecraft.core.Direction
import net.minecraft.core.Vec3i
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.IntTag
import net.minecraft.network.chat.Component
import net.minecraft.network.chat.TranslatableComponent
import net.minecraft.server.level.ServerLevel
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.state.BlockState
import net.minecraftforge.common.ForgeHooks
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.common.util.LazyOptional
import net.minecraftforge.energy.CapabilityEnergy
import net.minecraftforge.energy.IEnergyStorage
import net.minecraftforge.items.CapabilityItemHandler
import ru.dbotthepony.mc.otm.OverdriveThatMatters
import ru.dbotthepony.mc.otm.Registry
import ru.dbotthepony.mc.otm.capability.IMatteryEnergyStorage
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.container.MatteryContainer
import ru.dbotthepony.mc.otm.core.Fraction
import ru.dbotthepony.mc.otm.menu.ChemicalGeneratorMenu
import ru.dbotthepony.mc.otm.*
import ru.dbotthepony.mc.otm.block.BlockMatteryRotatable
import ru.dbotthepony.mc.otm.capability.receiveEnergy
import java.lang.ref.WeakReference

class BlockEntityChemicalGenerator(pos: BlockPos, state: BlockState) : BlockEntityMattery(Registry.BlockEntities.CHEMICAL_GENERATOR, pos, state) {
	override fun getDefaultDisplayName(): Component {
		return NAME
	}

	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu? {
		return ChemicalGeneratorMenu(containerID, inventory, this)
	}

	private var valid = true
	private var resolver = LazyOptional.of {energy}

	override fun <T> getCapability(cap: Capability<T>, side: Direction?): LazyOptional<T> {
		if (valid && (cap == MatteryCapability.ENERGY || cap == CapabilityEnergy.ENERGY) && side != blockState.getValue(BlockMatteryRotatable.FACING))
			return resolver.cast()

		if (valid && cap == CapabilityItemHandler.ITEM_HANDLER_CAPABILITY)
			return container_handler.get().cast()

		return super.getCapability(cap, side)
	}

	override fun invalidateCaps() {
		super.invalidateCaps()
		container_handler.invalidate()
		valid = false
		resolver.invalidate()
	}

	override fun reviveCaps() {
		super.reviveCaps()
		container_handler.revive()
		valid = true
		resolver = LazyOptional.of {energy}
	}

	override fun setLevel(level: Level) {
		super.setLevel(level)

		if (level is ServerLevel) {
			OverdriveThatMatters.tickOnce(level, this::checkSurroundings)
		}
	}

	companion object {
		private val THROUGHPUT = Fraction(160)
		private val THROUGHPUT_INT = 160
		private val GENERATION_SPEED = Fraction(40)
		private val MAX_ENERGY = Fraction(24_000)
		private val NAME = TranslatableComponent("block.overdrive_that_matters.chemical_generator")
	}

	override fun saveAdditional(nbt: CompoundTag) {
		super.saveAdditional(nbt)

		nbt["energy"] = power.serializeNBT()
		nbt["slots"] = container.serializeNBT()
		nbt["working_ticks"] = working_ticks
		nbt["working_ticks_total"] = working_ticks_total
	}

	override fun load(nbt: CompoundTag) {
		super.load(nbt)

		nbt.ifHas("energy") {
			power = Fraction.deserializeNBT(it)
		}

		nbt.ifHas("slots") {
			container.deserializeNBT(it)
		}

		nbt.ifHas("working_ticks") {
			if (it is IntTag)
				working_ticks = it.asInt
		}

		nbt.ifHas("working_ticks_total") {
			if (it is IntTag)
				working_ticks_total = it.asInt
		}
	}

	private val consumers = ArrayList<LazyOptional<IEnergyStorage>>()

	fun checkSurroundings() {
		if (!valid)
			return

		val known = consumers.clone() as ArrayList<*>
		consumers.clear()
		val level = level ?: return

		for (direction in Direction.values()) {
			// нельзя выталкивать энергию через перед
			if (direction == blockState.getValue(BlockMatteryRotatable.FACING))
				continue

			val resolver = level.getBlockEntity(blockPos + direction)?.getCapability(CapabilityEnergy.ENERGY, -direction)

			resolver?.ifPresent {
				if (!known.contains(resolver)) {
					val ref = WeakReference(this)

					resolver.addListener {
						ref.get()?.checkSurroundings()
					}
				}

				consumers.add(resolver)
			}
		}
	}

	val container = MatteryContainer({
		setChangedLight()
		check = true
	}, 2)

	val container_handler = container.handler(fun (slot, stack): Boolean {
		if (slot == 0)
			return ForgeHooks.getBurnTime(stack, null) > 0

		return stack.getCapability(CapabilityEnergy.ENERGY).isPresent
	}, fun (slot, amount, stack): Boolean {
		return slot == 1 &&
				(!stack.getCapability(CapabilityEnergy.ENERGY).isPresent || stack.getCapability(CapabilityEnergy.ENERGY).resolve().get().receiveEnergy(Int.MAX_VALUE, true) <= 0)
	})

	override fun setBlockState(p_155251_: BlockState) {
		super.setBlockState(p_155251_)

		if (valid && level is ServerLevel) {
			OverdriveThatMatters.tickOnce(level!!, this::checkSurroundings)
		}
	}

	private var power = Fraction.ZERO

	val energy = object : IMatteryEnergyStorage {
		override fun extractEnergyOuter(howMuch: Fraction, simulate: Boolean): Fraction {
			return extractEnergyInner(howMuch, simulate)
		}

		override fun extractEnergyInner(howMuch: Fraction, simulate: Boolean): Fraction {
			val extracted = howMuch.min(THROUGHPUT, power)

			if (!simulate && !extracted.isZero()) {
				power -= extracted
				check = true
				setChangedLight()
			}

			return extracted
		}

		override fun receiveEnergyOuter(howMuch: Fraction, simulate: Boolean): Fraction {
			return Fraction.ZERO
		}

		override fun receiveEnergyInner(howMuch: Fraction, simulate: Boolean): Fraction {
			val new_energy = (howMuch + power).min(MAX_ENERGY)
			val diff = new_energy - power

			if (!simulate) {
				power = new_energy
				setChangedLight()
			}

			return diff
		}

		override fun getBatteryLevel(): Fraction = power
		override fun getMaxBatteryLevel(): Fraction = MAX_ENERGY
		override fun canExtract(): Boolean = true
	}

	var working_ticks = 0
		private set

	var working_ticks_total = 0
		private set

	private var check = false

	fun workWithPower(it: IEnergyStorage) {
		if (it is IMatteryEnergyStorage) {
			val demand = it.missingPower
			val extracted = energy.extractEnergyInner(demand, true)
			val received = it.receiveEnergyOuter(extracted, false)

			if (!received.isZero()) {
				energy.extractEnergyInner(received, false)
			}
		} else {
			val demand = it.receiveEnergy(THROUGHPUT_INT, true)
			val extracted = energy.extractEnergyInner(demand, true)
			val received = it.receiveEnergy(extracted, false)

			if (received != 0) {
				energy.extractEnergyInner(received, false)
			}
		}
	}

	fun tick() {
		if (working_ticks > 0 && !isBlockedByRedstone) {
			working_ticks--
			energy.receiveEnergyInner(GENERATION_SPEED, false)

			if (working_ticks == 0) {
				working_ticks_total = 0
				check = true
			}
		}

		if (working_ticks == 0 && !isBlockedByRedstone && check) {
			if (!container.getItem(0).isEmpty) {
				val ticks = ForgeHooks.getBurnTime(container.getItem(0), null)

				if (ticks >= 4 && (energy.batteryLevel < Fraction.ONE || GENERATION_SPEED * (ticks / 4) + energy.batteryLevel < energy.maxBatteryLevel)) {
					working_ticks_total = ticks / 4
					working_ticks = ticks / 4
					container.getItem(0).shrink(1)
				}
			}

			check = false
		}

		if (power.isZero()) return

		val item = container.getItem(1)

		if (!item.isEmpty) {
			item.getCapability(CapabilityEnergy.ENERGY).ifPresent(this::workWithPower)
			if (power.isZero()) return
		}

		for (consumer in consumers) {
			consumer.ifPresent(this::workWithPower)
		}
	}
}
